@@ -67,43 +67,49 @@ static inline void batch_emit_float(struct sna *sna, float f)
 	batch_emit(sna, u.dw);
 }
 
-static inline Bool
+static inline bool
 is_gpu(DrawablePtr drawable)
 {
 	struct sna_pixmap *priv = sna_pixmap_from_drawable(drawable);
 
-	if (priv == NULL)
+	if (priv == NULL || priv->clear)
 		return false;
 
-	if (priv->gpu_damage)
+	if (DAMAGE_IS_ALL(priv->gpu_damage) || (priv->gpu_bo && kgem_bo_is_busy(priv->gpu_bo) && !priv->gpu_bo->proxy))
 		return true;
 
 	return priv->cpu_bo && kgem_bo_is_busy(priv->cpu_bo);
 }
 
-static inline Bool
+static inline bool
 is_cpu(DrawablePtr drawable)
 {
 	struct sna_pixmap *priv = sna_pixmap_from_drawable(drawable);
-	return !priv || priv->cpu_damage != NULL;
+	if (priv == NULL || priv->gpu_bo == NULL || priv->clear)
+		return true;
+
+	if (priv->gpu_damage && kgem_bo_is_busy(priv->gpu_bo))
+		return false;
+
+	if (priv->cpu_bo && kgem_bo_is_busy(priv->cpu_bo))
+		return false;
+
+	return true;
 }
 
-static inline Bool
+static inline bool
 is_dirty(DrawablePtr drawable)
 {
 	struct sna_pixmap *priv = sna_pixmap_from_drawable(drawable);
 	return priv == NULL || kgem_bo_is_dirty(priv->gpu_bo);
 }
 
-static inline Bool
-too_small(DrawablePtr drawable)
+static inline bool
+too_small(struct sna_pixmap *priv)
 {
-	struct sna_pixmap *priv = sna_pixmap_from_drawable(drawable);
+	assert(priv);
 
-	if (priv == NULL)
-		return true;
-
-	if (priv->gpu_damage)
+	if (priv->gpu_bo)
 		return false;
 
 	if (priv->cpu_bo && kgem_bo_is_busy(priv->cpu_bo))
@@ -112,40 +118,36 @@ too_small(DrawablePtr drawable)
 	return (priv->create & KGEM_CAN_CREATE_GPU) == 0;
 }
 
-static inline Bool
+static inline bool
 unattached(DrawablePtr drawable)
 {
 	struct sna_pixmap *priv = sna_pixmap_from_drawable(drawable);
-
-	if (priv == NULL)
-		return true;
-
-	return priv->gpu_bo == NULL && priv->cpu_bo == NULL;
+	return priv == NULL || (priv->gpu_damage == NULL && priv->cpu_damage);
 }
 
-static inline Bool
+static inline bool
 picture_is_gpu(PicturePtr picture)
 {
 	if (!picture || !picture->pDrawable)
-		return FALSE;
+		return false;
 	return is_gpu(picture->pDrawable);
 }
 
-static inline Bool sna_blt_compare_depth(DrawablePtr src, DrawablePtr dst)
+static inline bool sna_blt_compare_depth(DrawablePtr src, DrawablePtr dst)
 {
 	if (src->depth == dst->depth)
-		return TRUE;
+		return true;
 
 	/* Also allow for the alpha to be discarded on a copy */
 	if (src->bitsPerPixel != dst->bitsPerPixel)
-		return FALSE;
+		return false;
 
 	if (dst->depth == 24 && src->depth == 32)
-		return TRUE;
+		return true;
 
 	/* Note that a depth-16 pixmap is r5g6b5, not x1r5g5b5. */
 
-	return FALSE;
+	return false;
 }
 
 static inline struct kgem_bo *
@@ -187,4 +189,29 @@ sna_render_reduce_damage(struct sna_composite_op *op,
 	}
 }
 
+inline static uint32_t
+color_convert(uint32_t pixel,
+	      uint32_t src_format,
+	      uint32_t dst_format)
+{
+	DBG(("%s: src=%08x [%08x]\n", __FUNCTION__, pixel, src_format));
+
+	if (src_format != dst_format) {
+		uint16_t red, green, blue, alpha;
+
+		if (!sna_get_rgba_from_pixel(pixel,
+					     &red, &green, &blue, &alpha,
+					     src_format))
+			return 0;
+
+		if (!sna_get_pixel_from_rgba(&pixel,
+					     red, green, blue, alpha,
+					     dst_format))
+			return 0;
+	}
+
+	DBG(("%s: dst=%08x [%08x]\n", __FUNCTION__, pixel, dst_format));
+	return pixel;
+}
+
 #endif /* SNA_RENDER_INLINE_H */
